"""Enables retrieval of train departure information from Trafikverket API."""

from __future__ import annotations

from datetime import datetime

from .const import (
    DATE_TIME_FORMAT,
    TRAIN_STOP_REQUIRED_FIELDS,
)
from .exceptions import (
    MultipleTrainStationsFound,
    NoTrainAnnouncementFound,
    NoTrainStationFound,
)
from .filters import FieldFilter, FieldSort, FilterOperation, OrFilter, SortOrder
from .helpers import (
    station_from_xml_node,
    train_stop_from_xml_node,
)
from .models import StationInfoModel, TrainStopModel
from .trafikverket import (
    TrafikverketBase,
)


class TrafikverketTrain(TrafikverketBase):
    """Class used to communicate with trafikverket's train api."""

    version = "1.9"

    async def async_get_train_station_from_signature(
        self, signature: str
    ) -> StationInfoModel:
        """Retrieve train station based on signature."""
        train_stations = await self._api.async_make_request(
            "TrainStation",
            "1.5",
            "rail.infrastructure",
            [
                "AdvertisedLocationName",
                "LocationSignature",
                "Advertised",
                "Deleted",
                "OfficialLocationName",
            ],
            [
                FieldFilter(FilterOperation.EQUAL, "LocationSignature", signature),
            ],
        )
        if len(train_stations) == 0:
            raise NoTrainStationFound(
                "Could not find a station with the specified name"
            )
        return await station_from_xml_node(train_stations[0])

    async def async_search_train_station(self, location_name: str) -> StationInfoModel:
        """Retrieve train station id based on name."""
        train_stations = await self.async_search_train_stations(location_name)

        if len(train_stations) == 0:
            raise NoTrainStationFound(
                "Could not find a station with the specified name"
            )
        if len(train_stations) > 1:
            raise MultipleTrainStationsFound(
                "Found multiple stations with the specified name"
            )

        return train_stations[0]

    async def async_search_train_stations(
        self, location_name: str
    ) -> list[StationInfoModel]:
        """Search for train stations."""
        train_stations = await self._api.async_make_request(
            "TrainStation",
            "1.5",
            "rail.infrastructure",
            [
                "AdvertisedLocationName",
                "LocationSignature",
                "Advertised",
                "Deleted",
                "OfficialLocationName",
            ],
            [
                OrFilter(
                    [
                        FieldFilter(
                            FilterOperation.LIKE,
                            "AdvertisedLocationName",
                            location_name,
                        ),
                        FieldFilter(
                            FilterOperation.LIKE,
                            "OfficialLocationName",
                            location_name,
                        ),
                    ]
                ),
                FieldFilter(FilterOperation.EQUAL, "Advertised", "true"),
            ],
        )
        if len(train_stations) == 0:
            raise NoTrainStationFound(
                "Could not find a station with the specified name"
            )

        result = []

        for train_station in train_stations:
            result.append(await station_from_xml_node(train_station))

        return result

    async def async_get_train_stop(
        self,
        from_station: StationInfoModel,
        to_station: StationInfoModel,
        time_at_location: datetime,
        product_description: str | None = None,
        exclude_canceled: bool = False,
    ) -> TrainStopModel:
        """Retrieve the train stop."""
        date_as_text = time_at_location.strftime(DATE_TIME_FORMAT)

        filters = [
            FieldFilter(FilterOperation.EQUAL, "ActivityType", "Avgang"),
            FieldFilter(
                FilterOperation.EQUAL, "LocationSignature", from_station.signature
            ),
            FieldFilter(
                FilterOperation.EQUAL, "AdvertisedTimeAtLocation", date_as_text
            ),
            OrFilter(
                [
                    FieldFilter(
                        FilterOperation.EQUAL,
                        "ViaToLocation.LocationName",
                        to_station.signature,
                    ),
                    FieldFilter(
                        FilterOperation.EQUAL,
                        "ToLocation.LocationName",
                        to_station.signature,
                    ),
                ]
            ),
        ]

        if product_description:
            filters.append(
                FieldFilter(
                    FilterOperation.LIKE,
                    "ProductInformation.Description",
                    product_description,
                )
            )

        if exclude_canceled:
            filters.append(FieldFilter(FilterOperation.EQUAL, "Canceled", "false"))

        train_announcements = await self._api.async_make_request(
            "TrainAnnouncement",
            self.version,
            None,
            TRAIN_STOP_REQUIRED_FIELDS,
            filters,
            1,
        )

        if len(train_announcements) == 0:
            raise NoTrainAnnouncementFound("No TrainAnnouncement found")

        train_announcement = train_announcements[0]
        return await train_stop_from_xml_node(train_announcement)

    async def async_get_next_train_stops(
        self,
        from_station: StationInfoModel,
        to_station: StationInfoModel,
        after_time: datetime,
        product_description: str | None = None,
        exclude_canceled: bool = False,
        number_of_stops: int = 1,
    ) -> list[TrainStopModel]:
        """Enable retrieval of next departures."""
        date_as_text = after_time.strftime(DATE_TIME_FORMAT)

        filters = [
            FieldFilter(FilterOperation.EQUAL, "ActivityType", "Avgang"),
            FieldFilter(
                FilterOperation.EQUAL, "LocationSignature", from_station.signature
            ),
            FieldFilter(
                FilterOperation.GREATER_THAN_EQUAL,
                "AdvertisedTimeAtLocation",
                date_as_text,
            ),
            OrFilter(
                [
                    FieldFilter(
                        FilterOperation.EQUAL,
                        "ViaToLocation.LocationName",
                        to_station.signature,
                    ),
                    FieldFilter(
                        FilterOperation.EQUAL,
                        "ToLocation.LocationName",
                        to_station.signature,
                    ),
                ]
            ),
        ]

        if product_description:
            filters.append(
                FieldFilter(
                    FilterOperation.LIKE,
                    "ProductInformation.Description",
                    product_description,
                )
            )

        if exclude_canceled:
            filters.append(FieldFilter(FilterOperation.EQUAL, "Canceled", "false"))

        sorting = [FieldSort("AdvertisedTimeAtLocation", SortOrder.ASCENDING)]
        train_announcements = await self._api.async_make_request(
            "TrainAnnouncement",
            self.version,
            None,
            TRAIN_STOP_REQUIRED_FIELDS,
            filters,
            number_of_stops,
            sorting,
        )

        if len(train_announcements) == 0:
            raise NoTrainAnnouncementFound("No TrainAnnouncement found")

        stops = []
        for announcement in train_announcements:
            stops.append(await train_stop_from_xml_node(announcement))
        return stops

    async def async_get_next_train_stop(
        self,
        from_station: StationInfoModel,
        to_station: StationInfoModel,
        after_time: datetime,
        product_description: str | None = None,
        exclude_canceled: bool = False,
    ) -> TrainStopModel:
        """Enable retrieval of next departure."""

        stops = await self.async_get_next_train_stops(
            from_station,
            to_station,
            after_time,
            product_description,
            exclude_canceled,
            1,
        )
        return stops[0]



"""Constants for the pytrafikverket library."""

from __future__ import annotations

API_URL = "https://api.trafikinfo.trafikverket.se/v2/data.xml"

SWEDEN_TIMEZONE = "Europe/Stockholm"
DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
DATE_TIME_FORMAT_FOR_MODIFIED = "%Y-%m-%dT%H:%M:%S.%fZ"

SEARCH_FOR_STATION = "search-for-station"
GET_TRAIN_STOP = "get-train-stop"
GET_NEXT_TRAIN_STOP = "get-next-train-stop"
GET_WEATHER = "get-weather"
GET_FERRY_ROUTE = "get-ferry-route"
SEARCH_FOR_FERRY_ROUTE = "search-for-ferry-route"
GET_NEXT_FERRY_STOP = "get-next-ferry-stop"
DATE_TIME_INPUT = "%Y-%m-%dT%H:%M:%S"

CAMERA_INFO_REQUIRED_FIELDS = [
    "Name",
    "Id",
    "Active",
    "Deleted",
    "Description",
    "Direction",
    "HasFullSizePhoto",
    "Location",
    "ModifiedTime",
    "PhotoTime",
    "PhotoUrl",
    "Status",
    "Type",
]
ROUTE_INFO_REQUIRED_FIELDS = ["Id", "Name", "Shortname", "Type.Name"]
DEVIATION_INFO_REQUIRED_FIELDS = [
    "Deviation.Id",
    "Deviation.Header",
    "Deviation.EndTime",
    "Deviation.StartTime",
    "Deviation.Message",
    "Deviation.IconId",
    "Deviation.LocationDescriptor",
]
FERRY_STOP_REQUIRED_FIELDS = [
    "Id",
    "Deleted",
    "DepartureTime",
    "Route.Name",
    "DeviationId",
    "ModifiedTime",
    "FromHarbor",
    "ToHarbor",
    "Info",
]
STATION_INFO_REQUIRED_FIELDS = ["LocationSignature", "AdvertisedLocationName"]
TRAIN_STOP_REQUIRED_FIELDS = [
    "ActivityId",
    "Canceled",
    "AdvertisedTimeAtLocation",
    "EstimatedTimeAtLocation",
    "TimeAtLocation",
    "OtherInformation",
    "Deviation",
    "ModifiedTime",
    "ProductInformation",
]
WEATHER_REQUIRED_FIELDS = [
    "Name",  # string, replaced
    "Id",  # string, replaced
    "ModifiedTime",  # datetime, new, Tidpunkt då dataposten ändrades i cachen
    "Observation.Sample",  # datetime, replaced, Tidpunkt som observationen avser, inklusive tidzon för att hantera sommartid och normaltid. # codespell:ignore
    "Observation.Air.Temperature.Value",  # float, replaced, Lufttemperatur. Value [C]
    "Observation.Air.RelativeHumidity.Value",  # float, replaced, Relativ luftfuktighet. Andel av den fukt som luften kan bära. Vid 100% är luften mättad. Value [%] # codespell:ignore
    "Observation.Air.Dewpoint.Value",  # float, new, Daggpunkt, den temperatur där vatten kondenserar. Value [C] # codespell:ignore
    "Observation.Air.VisibleDistance.Value",  # float, new, Den sträcka det finns sikt. Value [m]
    "Observation.Wind.Direction.Value",  # int, replaced, Mått på vindriktning vid en viss tidpunkt. Medelvärde över tiominutersperiod t.o.m. tidpunkten. Value [grader]
    "Observation.Wind.Height",  # int, new, Vindsensorns höjdplacering [m]
    "Observation.Wind.Speed.Value",  # float, replaced, Mått på vindhastighet vid en viss tidpunkt. Medelvärde över tiominutersperiod t.o.m. tidpunkten. Value [m/s]
    "Observation.Aggregated30minutes.Wind.SpeedMax.Value",  # float, replaced, Högst uppmätt 3-sekundersmedelvärde under perioden. Value [m/s]
    "Observation.Weather.Precipitation",  # string, replaced, Vilken typ av nederbörd som detekterats # codespell:ignore
    "Observation.Aggregated30minutes.Precipitation.TotalWaterEquivalent.Value",  # float, replaced, Mängd vatten som nederbörden under perioden motsvarar. Value [mm] # codespell:ignore
    "Observation.Aggregated30minutes.Precipitation.Rain",  # bool, new, Förekomst av regn.
    "Observation.Aggregated30minutes.Precipitation.Snow",  # bool, new, Förekomst av snö.
    "Observation.Surface.Temperature.Value",  # float, replaced, Vägytans temperatur. Value [C] # codespell:ignore
    "Observation.Surface.Ice",  # bool, new, Förekomst av is på vägytan.
    "Observation.Surface.IceDepth.Value",  # float, new, Isdjup på vägytan. Value [mm]
    "Observation.Surface.Snow",  # bool, new, Förekomst av snö på vägytan.
    "Observation.Surface.SnowDepth.Solid.Value",  # float, new, Mängd snö. Value [mm]
    "Observation.Surface.SnowDepth.WaterEquivalent.Value",  # float, new, Mängd vatten som snön motsvarar i smält form. Value [mm] # codespell:ignore
    "Observation.Surface.Water",  # bool, new, Förekomst av vatten på vägytan.
    "Observation.Surface.WaterDepth.Value",  # float, new, Vattendjup på vägytan. Value [mm]
]  # Precipitation possible values are: no, rain, freezing_rain, snow, sleet, yes



"""Helpers for creating dataclasses."""

from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING

import aiozoneinfo
from lxml import etree

from .const import DATE_TIME_FORMAT, DATE_TIME_FORMAT_FOR_MODIFIED, SWEDEN_TIMEZONE
from .models import (
    CameraInfoModel,
    DeviationInfoModel,
    FerryRouteInfoModel,
    FerryStopModel,
    StationInfoModel,
    TrainStopModel,
    WeatherStationInfoModel,
)

LOGGER = logging.getLogger(__name__)


class NodeHelper:
    """Helper class to get node content."""

    def __init__(self, node: etree._ElementTree) -> None:
        """Initialize the class."""
        self._node = node

    def get_text(self, field: str) -> str | None:
        """Return the text in 'field' from the node or None if not found."""
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        if len(nodes) == 0:
            return None
        if len(nodes) > 1:
            raise ValueError("Found multiple nodes should only 0 or 1 is allowed")
        LOGGER.debug("Return text value %s", nodes[0].text)
        value: str | None = nodes[0].text
        return value

    def get_number(self, field: str) -> float | None:
        """Return the number in 'field' from the node or None if not found."""
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        if len(nodes) == 0:
            return None
        if len(nodes) > 1:
            raise ValueError("Found multiple nodes should only 0 or 1 is allowed")
        LOGGER.debug("Return number value %s", nodes[0].text)
        if TYPE_CHECKING:
            assert nodes[0].text is not None
        try:
            value = float(nodes[0].text)
        except ValueError:
            return None
        return value

    def get_texts(self, field: str) -> list[str] | None:
        """Return a list of texts from the node selected by 'field' or None."""
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        result: list[str] = []
        for line in nodes:
            if TYPE_CHECKING:
                assert line.text is not None
            result.append(line.text)
        LOGGER.debug("Return texts value %s", result)
        return result

    def get_datetime_for_modified(self, field: str) -> datetime.datetime | None:
        """Return datetime object from node, selected by 'field'.

        Format of the text is expected to be modifiedTime-format.
        """
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        if len(nodes) == 0:
            return None
        if len(nodes) > 1:
            raise ValueError("Found multiple nodes should only 0 or 1 is allowed")
        LOGGER.debug("Return modified value %s", nodes[0].text)
        if TYPE_CHECKING:
            assert nodes[0].text is not None
        return datetime.datetime.strptime(nodes[0].text, DATE_TIME_FORMAT_FOR_MODIFIED)

    def get_datetime(self, field: str) -> datetime.datetime | None:
        """Return a datetime object from node, selected by 'field'."""
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        if len(nodes) == 0:
            return None
        if len(nodes) > 1:
            raise ValueError("Found multiple nodes should only 0 or 1 is allowed")
        LOGGER.debug("Return datetime value %s", nodes[0].text)
        if TYPE_CHECKING:
            assert nodes[0].text is not None
        return datetime.datetime.strptime(nodes[0].text, DATE_TIME_FORMAT)

    def get_bool(self, field: str) -> bool | None:
        """Return True if value selected by field is 'true' else returns False."""
        nodes: list[etree._Element] | None = self._node.xpath(field)
        if nodes is None:
            return None
        if len(nodes) == 0:
            return None
        if len(nodes) > 1:
            raise ValueError("Found multiple nodes should only 0 or 1 is allowed")
        LOGGER.debug("Return bool value %s", nodes[0].text)
        if TYPE_CHECKING:
            assert nodes[0].text is not None
        value: bool = nodes[0].text.lower() == "true"
        return value


async def weather_from_xml_node(node: etree._ElementTree) -> WeatherStationInfoModel:
    """Map XML path for values."""
    node_helper = NodeHelper(node)
    station_name = node_helper.get_text("Name")
    station_id = node_helper.get_text("Id")
    air_temp = node_helper.get_number("Observation/Air/Temperature/Value")
    road_temp = node_helper.get_number("Observation/Surface/Temperature/Value")
    dew_point = node_helper.get_number("Observation/Air/Dewpoint/Value")
    humidity = node_helper.get_number("Observation/Air/RelativeHumidity/Value")
    visible_distance = node_helper.get_number("Observation/Air/VisibleDistance/Value")
    precipitationtype = node_helper.get_text("Observation/Weather/Precipitation")
    raining = node_helper.get_bool("Observation/Aggregated30minutes/Precipitation/Rain")
    snowing = node_helper.get_bool("Observation/Aggregated30minutes/Precipitation/Snow")
    road_ice = node_helper.get_bool("Observation/Surface/Ice")
    road_ice_depth = node_helper.get_number("Observation/Surface/IceDepth/Value")
    road_snow = node_helper.get_bool("Observation/Surface/Snow")
    road_snow_depth = node_helper.get_number(
        "Observation/Surface/SnowDepth.Solid/Value"
    )
    road_water = node_helper.get_bool("Observation/Surface/Water")
    road_water_depth = node_helper.get_number("Observation/Surface/WaterDepth/Value")
    road_water_equivalent_depth = node_helper.get_number(
        "Observation/Surface/SnowDepth/WaterEquivalent/Value"
    )
    winddirection = node_helper.get_text("Observation/Wind/Direction/Value")
    wind_height = node_helper.get_number("Observation/Wind/Height")
    windforce = node_helper.get_number("Observation/Wind/Speed/Value")
    windforcemax = node_helper.get_number(
        "Observation/Aggregated30minutes/Wind/SpeedMax/Value"
    )
    measure_time = node_helper.get_datetime("Observation/Sample")
    precipitation_amount = node_helper.get_number(
        "Observation/Aggregated30minutes/Precipitation/TotalWaterEquivalent/Value"
    )
    if precipitation_amount:
        precipitation_amount = precipitation_amount * 2  # mm/30min to mm/h
    modified_time = node_helper.get_datetime_for_modified("ModifiedTime")

    zoneinfo = await aiozoneinfo.async_get_time_zone(SWEDEN_TIMEZONE)
    if measure_time:
        measure_time = measure_time.replace(tzinfo=zoneinfo)
    if modified_time:
        modified_time = modified_time.replace(tzinfo=datetime.UTC)

    if TYPE_CHECKING:
        assert station_name
        assert station_id

    return WeatherStationInfoModel(
        station_name=station_name,
        station_id=station_id,
        road_temp=road_temp,
        air_temp=air_temp,
        dew_point=dew_point,
        humidity=humidity,
        visible_distance=visible_distance,
        precipitationtype=precipitationtype,
        raining=raining,
        snowing=snowing,
        road_ice=road_ice,
        road_ice_depth=road_ice_depth,
        road_snow=road_snow,
        road_snow_depth=road_snow_depth,
        road_water=road_water,
        road_water_depth=road_water_depth,
        road_water_equivalent_depth=road_water_equivalent_depth,
        winddirection=winddirection,
        wind_height=wind_height,
        windforce=windforce,
        windforcemax=windforcemax,
        measure_time=measure_time,
        precipitation_amount=precipitation_amount,
        modified_time=modified_time,
    )


async def camera_from_xml_node(node: etree._ElementTree) -> CameraInfoModel:
    """Map XML path for values."""
    node_helper = NodeHelper(node)
    camera_name = node_helper.get_text("Name")
    camera_id = node_helper.get_text("Id")
    active = node_helper.get_bool("Active")
    deleted = node_helper.get_bool("Deleted")
    description = node_helper.get_text("Description")
    direction = node_helper.get_text("Direction")
    fullsizephoto = node_helper.get_bool("HasFullSizePhoto")
    location = node_helper.get_text("Location")
    modified = node_helper.get_datetime_for_modified("ModifiedTime")
    phototime = node_helper.get_datetime("PhotoTime")
    photourl = node_helper.get_text("PhotoUrl")
    status = node_helper.get_text("Status")
    camera_type = node_helper.get_text("Type")

    zoneinfo = await aiozoneinfo.async_get_time_zone(SWEDEN_TIMEZONE)
    if phototime:
        phototime = phototime.replace(tzinfo=zoneinfo)
    if modified:
        modified = modified.replace(tzinfo=datetime.UTC)

    if TYPE_CHECKING:
        assert camera_name
        assert camera_id

    return CameraInfoModel(
        camera_name=camera_name,
        camera_id=camera_id,
        active=active,
        deleted=deleted,
        description=description,
        direction=direction,
        fullsizephoto=fullsizephoto,
        location=location,
        modified=modified,
        phototime=phototime,
        photourl=photourl,
        status=status,
        camera_type=camera_type,
    )


async def ferry_stop_from_xml_node(node: etree._ElementTree) -> FerryStopModel:
    """Map the path in the return XML data."""
    node_helper = NodeHelper(node)
    id = node_helper.get_text("Id")
    name = node_helper.get_text("Route/Name")
    short_name = node_helper.get_text("Route/Shortname")
    deleted = node_helper.get_bool("Deleted")
    departure_time = node_helper.get_datetime("DepartureTime")
    other_information = node_helper.get_texts("Info")
    deviation_id = node_helper.get_texts("DeviationId")
    modified_time = node_helper.get_datetime_for_modified("ModifiedTime")
    from_harbor_name = node_helper.get_text("FromHarbor/Name")
    to_harbor_name = node_helper.get_text("ToHarbor/Name")
    type_name = node_helper.get_text("Route/Type/Name")

    zoneinfo = await aiozoneinfo.async_get_time_zone(SWEDEN_TIMEZONE)
    if departure_time:
        departure_time = departure_time.replace(tzinfo=zoneinfo)
    if modified_time:
        modified_time = modified_time.replace(tzinfo=datetime.UTC)

    if TYPE_CHECKING:
        assert id
        assert name

    return FerryStopModel(
        ferry_stop_id=id,
        ferry_stop_name=name,
        short_name=short_name,
        deleted=deleted,
        departure_time=departure_time,
        other_information=other_information,
        deviation_id=deviation_id,
        modified_time=modified_time,
        from_harbor_name=from_harbor_name,
        to_harbor_name=to_harbor_name,
        type_name=type_name,
    )


async def deviation_from_xml_node(node: etree._ElementTree) -> DeviationInfoModel:
    """Map deviation information in XML data."""
    node_helper = NodeHelper(node)
    id = node_helper.get_text("Deviation/Id")
    header = node_helper.get_text("Deviation/Header")
    message = node_helper.get_text("Deviation/Message")
    start_time = node_helper.get_datetime("Deviation/StartTime")
    end_time = node_helper.get_datetime("Deviation/EndTime")
    icon_id = node_helper.get_text("Deviation/IconId")
    location_desc = node_helper.get_text("Deviation/LocationDescriptor")

    zoneinfo = await aiozoneinfo.async_get_time_zone(SWEDEN_TIMEZONE)
    if start_time:
        start_time = start_time.replace(tzinfo=zoneinfo)
    if end_time:
        end_time = end_time.replace(tzinfo=zoneinfo)

    if TYPE_CHECKING:
        assert id

    return DeviationInfoModel(
        deviation_id=id,
        header=header,
        message=message,
        start_time=start_time,
        end_time=end_time,
        icon_id=icon_id,
        location_desc=location_desc,
    )


async def ferry_route_from_xml_node(node: etree._ElementTree) -> FerryRouteInfoModel:
    """Map route information in XML data."""
    node_helper = NodeHelper(node)
    id = node_helper.get_text("Id")
    name = node_helper.get_text("Name")
    short_name = node_helper.get_text("Shortname")
    route_type = node_helper.get_text("Type/Name")

    if TYPE_CHECKING:
        assert id
        assert name

    return FerryRouteInfoModel(
        ferry_route_id=id,
        ferry_route_name=name,
        short_name=short_name,
        route_type=route_type,
    )


async def station_from_xml_node(node: etree._ElementTree) -> StationInfoModel:
    """Map station information in XML data."""
    node_helper = NodeHelper(node)
    location_signature = node_helper.get_text("LocationSignature")
    advertised_location_name = node_helper.get_text("AdvertisedLocationName")
    location_advertised = node_helper.get_text("Advertised")

    if TYPE_CHECKING:
        assert location_signature
        assert advertised_location_name

    return StationInfoModel(
        signature=location_signature,
        station_name=advertised_location_name,
        advertised=location_advertised,
    )


async def train_stop_from_xml_node(node: etree._ElementTree) -> TrainStopModel:
    """Map the path in the return XML data."""
    node_helper = NodeHelper(node)
    activity_id = node_helper.get_text("ActivityId")
    canceled = node_helper.get_bool("Canceled")
    advertised_time_at_location = node_helper.get_datetime("AdvertisedTimeAtLocation")
    estimated_time_at_location = node_helper.get_datetime("EstimatedTimeAtLocation")
    time_at_location = node_helper.get_datetime("TimeAtLocation")
    other_information = node_helper.get_texts("OtherInformation/Description")
    deviations = node_helper.get_texts("Deviation/Description")
    modified_time = node_helper.get_datetime_for_modified("ModifiedTime")
    product_description = node_helper.get_texts("ProductInformation/Description")

    zoneinfo = await aiozoneinfo.async_get_time_zone(SWEDEN_TIMEZONE)
    if advertised_time_at_location:
        advertised_time_at_location = advertised_time_at_location.replace(
            tzinfo=zoneinfo
        )
    if estimated_time_at_location:
        estimated_time_at_location = estimated_time_at_location.replace(tzinfo=zoneinfo)
    if time_at_location:
        time_at_location = time_at_location.replace(tzinfo=zoneinfo)
    if modified_time:
        modified_time = modified_time.replace(tzinfo=datetime.UTC)

    if TYPE_CHECKING:
        assert activity_id

    return TrainStopModel(
        train_stop_id=activity_id,
        canceled=canceled,
        advertised_time_at_location=advertised_time_at_location,
        estimated_time_at_location=estimated_time_at_location,
        time_at_location=time_at_location,
        other_information=other_information,
        deviations=deviations,
        modified_time=modified_time,
        product_description=product_description,
    )



"""CLI enabler for pytrafikverket."""

from __future__ import annotations

import argparse
import asyncio
import json
from datetime import datetime
from typing import Any

import aiohttp

from .const import (
    DATE_TIME_INPUT,
    GET_FERRY_ROUTE,
    GET_NEXT_FERRY_STOP,
    GET_NEXT_TRAIN_STOP,
    GET_TRAIN_STOP,
    GET_WEATHER,
    SEARCH_FOR_FERRY_ROUTE,
    SEARCH_FOR_STATION,
)
from .trafikverket_ferry import TrafikverketFerry
from .trafikverket_train import TrafikverketTrain
from .trafikverket_weather import TrafikverketWeather


async def async_main(loop: Any) -> None:
    """Set up function to handle input and get data to present."""
    async with aiohttp.ClientSession(loop=loop) as session:
        parser = argparse.ArgumentParser(
            description="CLI used to get data from trafikverket"
        )
        parser.add_argument("-key", type=str, required=True)
        parser.add_argument(
            "-method",
            required=True,
            choices=(
                SEARCH_FOR_STATION,
                GET_TRAIN_STOP,
                GET_NEXT_TRAIN_STOP,
                GET_WEATHER,
                GET_FERRY_ROUTE,
                SEARCH_FOR_FERRY_ROUTE,
                GET_NEXT_FERRY_STOP,
            ),
        )
        parser.add_argument("-station", type=str)
        parser.add_argument("-from-station", type=str)
        parser.add_argument("-to-station", type=str)
        parser.add_argument("-date-time", type=str)
        parser.add_argument("-route", type=str)
        parser.add_argument("-from-harbor", type=str)
        parser.add_argument("-to-harbor", type=str)
        parser.add_argument("-train-product", type=str)
        parser.add_argument(
            "-exclude-canceled-trains",
            action=argparse.BooleanOptionalAction,
            default=False,
        )

        async with asyncio.timeout(10):
            args = parser.parse_args()
            train_api = TrafikverketTrain(session, args.key)
            weather_api = TrafikverketWeather(session, args.key)
            ferry_api = TrafikverketFerry(session, args.key)

            if args.method == SEARCH_FOR_STATION:
                if args.station is None:
                    raise ValueError("-station is required")
                stations = await train_api.async_search_train_stations(args.station)
                for station in stations:
                    print_values(station)
            elif args.method == GET_TRAIN_STOP:
                from_station = await train_api.async_search_train_station(
                    args.from_station
                )
                to_station = await train_api.async_search_train_station(args.to_station)
                print(f"from_station_signature: {from_station.signature}")  # noqa: T201
                print(f"to_station_signature:   {to_station.signature}")  # noqa: T201

                time = datetime.strptime(args.date_time, DATE_TIME_INPUT)

                train_stop = await train_api.async_get_train_stop(
                    from_station,
                    to_station,
                    time,
                    product_description=args.train_product,
                    exclude_canceled=args.exclude_canceled_trains,
                )
                print_values(train_stop)

            elif args.method == GET_NEXT_TRAIN_STOP:
                from_station = await train_api.async_search_train_station(
                    args.from_station
                )
                to_station = await train_api.async_search_train_station(args.to_station)
                print(f"from_station_signature: {from_station.signature}")  # noqa: T201
                print(f"to_station_signature:   {to_station.signature}")  # noqa: T201

                if args.date_time is not None:
                    time = datetime.strptime(args.date_time, DATE_TIME_INPUT)
                else:
                    time = datetime.now()
                train_stop = await train_api.async_get_next_train_stop(
                    from_station,
                    to_station,
                    time,
                    product_description=args.train_product,
                    exclude_canceled=args.exclude_canceled_trains,
                )
                print_values(train_stop)

            elif args.method == GET_WEATHER:
                if args.station is None:
                    raise ValueError(
                        '-station is required with name of Weather station\
                         (ex. -station "Nöbbele")'
                    )
                weather = await weather_api.async_get_weather(args.station)
                print_values(weather)

            elif args.method == GET_FERRY_ROUTE:
                if args.route is None:
                    raise ValueError(
                        '-route is required with name of Ferry route\
                         (ex. -route "Ekeröleden")'
                    )
                route = await ferry_api.async_get_ferry_route(args.route)
                print_values(route)

            elif args.method == SEARCH_FOR_FERRY_ROUTE:
                if args.route is None:
                    raise ValueError("-route is required")
                routes = await ferry_api.async_search_ferry_routes(args.route)
                for route in routes:
                    print_values(route)
                    print(f"{route.ferry_route_name} {route.ferry_route_id}")  # noqa: T201

            elif args.method == GET_NEXT_FERRY_STOP:
                if args.from_harbor is None:
                    raise ValueError(
                        '-from-harbor is required with name of Ferry harbor\
                         (ex. -from-harbor "Ekerö")'
                    )
                if args.date_time is not None:
                    time = datetime.strptime(args.date_time, DATE_TIME_INPUT)
                else:
                    time = datetime.now()

                ferry_stop = await ferry_api.async_get_next_ferry_stop(
                    args.from_harbor, args.to_harbor, time
                )
                print_values(ferry_stop)


def print_values(result: Any) -> None:
    """Print out values for all object members."""
    print(json.dumps(result.__dict__, indent=4, ensure_ascii=False))  # noqa: T201


def main() -> None:
    """Initialize the loop."""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(async_main(loop))



"""Exceptions for Trafikverket."""

from __future__ import annotations


class NoCameraFound(Exception):
    """Error found no camera."""


class MultipleCamerasFound(Exception):
    """Error found multiple cameras."""


class NoRouteFound(Exception):
    """Error found no ferry route."""


class MultipleRoutesFound(Exception):
    """Error found multiple ferry routes."""


class NoFerryFound(Exception):
    """Error found no ferry."""


class NoDeviationFound(Exception):
    """Error found no deviation."""


class MultipleDeviationsFound(Exception):
    """Error found multiple deviations."""


class NoWeatherStationFound(Exception):
    """Error found no weather station."""


class MultipleWeatherStationsFound(Exception):
    """Error found multiple weather stations."""


class NoTrainStationFound(Exception):
    """Error found no train station."""


class MultipleTrainStationsFound(Exception):
    """Error found multiple train stations."""


class NoTrainAnnouncementFound(Exception):
    """Error found no train announcement."""


class InvalidAuthentication(Exception):
    """Error found multiple train announcements."""


class UnknownError(Exception):
    """Error found multiple train announcements."""



"""Filters and sorting for Trafikverket API requests."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from enum import StrEnum
from typing import Any

from lxml import etree


class FilterOperation(StrEnum):
    """Contains all field filter operations."""

    EQUAL = "EQ"
    EXISTS = "EXISTS"
    GREATER_THAN = "GT"
    GREATER_THAN_EQUAL = "GTE"
    LESS_THAN = "LT"
    LESS_THAN_EQUAL = "LTE"
    NOT_EQUAL = "NE"
    LIKE = "LIKE"
    NOT_LIKE = "NOTLIKE"
    #    IN = "IN"  # noqa: ERA001
    NOT_IN = "NOTIN"  # codespell:ignore
    WITH_IN = "WITHIN"


class SortOrder(StrEnum):
    """Specifies how rows of data are sorted."""

    ASCENDING = "asc"
    DESCENDING = "desc"


class FieldSort:
    """What field and how to sort on it."""

    def __init__(self, field: str, sort_order: SortOrder) -> None:
        """Initialize the class."""
        self._field = field
        self._sort_order = sort_order

    def to_string(self) -> str:
        """Sort_order as string."""
        return self._field + " " + self._sort_order.value


class Filter:
    """Base class for all filters."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def generate_node(self, parent_node: etree._Element) -> Any:
        """Generate node."""


class FieldFilter(Filter):
    """Used to filter on one field."""

    def __init__(self, operation: FilterOperation, name: str, value: Any) -> None:
        """Initialize the class."""
        self.operation = operation
        self.name = name
        self.value = value

    def generate_node(self, parent_node: etree._Element) -> Any:
        """Return element node for field filter."""
        filter_node: etree._Element = etree.SubElement(
            parent_node, self.operation.value
        )
        filter_node.attrib["name"] = self.name
        filter_node.attrib["value"] = self.value
        return filter_node


class OrFilter(Filter):
    """Used to create a Or filter."""

    def __init__(self, filters: list[Filter]) -> None:
        """Initialize the class."""
        self.filters = filters

    def generate_node(self, parent_node: Any) -> Any:
        """Return element node for filter."""
        or_node: etree._Element = etree.SubElement(parent_node, "OR")
        for sub_filter in self.filters:
            sub_filter.generate_node(or_node)
        return or_node


class AndFilter(Filter):
    """Used to create a And filter."""

    def __init__(self, filters: list[Filter]) -> None:
        """Initialize the class."""
        self.filters = filters

    def generate_node(self, parent_node: Any) -> etree._Element:
        """Return element node for filter."""
        or_node: etree._Element = etree.SubElement(parent_node, "AND")
        for sub_filter in self.filters:
            sub_filter.generate_node(or_node)
        return or_node



"""Enables retrieval of ferry departure information from Trafikverket API."""

from __future__ import annotations

from datetime import datetime

from .const import (
    DATE_TIME_FORMAT,
    DEVIATION_INFO_REQUIRED_FIELDS,
    FERRY_STOP_REQUIRED_FIELDS,
    ROUTE_INFO_REQUIRED_FIELDS,
)
from .exceptions import (
    MultipleDeviationsFound,
    MultipleRoutesFound,
    NoDeviationFound,
    NoFerryFound,
    NoRouteFound,
)
from .filters import FieldFilter, FieldSort, Filter, FilterOperation, SortOrder
from .helpers import (
    deviation_from_xml_node,
    ferry_route_from_xml_node,
    ferry_stop_from_xml_node,
)
from .models import (
    DeviationInfoModel,
    FerryRouteInfoModel,
    FerryStopModel,
)
from .trafikverket import (
    TrafikverketBase,
)


class TrafikverketFerry(TrafikverketBase):
    """Class used to communicate with trafikverket's ferry route api."""

    version = "1.2"

    async def async_get_ferry_route(self, route_name: str) -> FerryRouteInfoModel:
        """Retrieve ferry route id based on name."""
        routes = await self._api.async_make_request(
            "FerryRoute",
            self.version,
            None,
            ROUTE_INFO_REQUIRED_FIELDS,
            [FieldFilter(FilterOperation.EQUAL, "Name", route_name)],
        )
        if len(routes) == 0:
            raise NoRouteFound("Could not find a route with the specified name")
        if len(routes) > 1:
            raise MultipleRoutesFound("Found multiple routes with the specified name")

        return await ferry_route_from_xml_node(routes[0])

    async def async_get_ferry_route_id(self, route_id: int) -> FerryRouteInfoModel:
        """Retrieve ferry route id based on routeId."""
        routes = await self._api.async_make_request(
            "FerryRoute",
            self.version,
            None,
            ROUTE_INFO_REQUIRED_FIELDS,
            [FieldFilter(FilterOperation.EQUAL, "Id", str(route_id))],
        )
        if len(routes) == 0:
            raise NoRouteFound("Could not find a route with the specified name")
        if len(routes) > 1:
            raise MultipleRoutesFound("Found multiple routes with the specified name")

        return await ferry_route_from_xml_node(routes[0])

    async def async_search_ferry_routes(self, name: str) -> list[FerryRouteInfoModel]:
        """Search for ferry routes based on the route name."""
        routes = await self._api.async_make_request(
            "FerryRoute",
            self.version,
            None,
            ROUTE_INFO_REQUIRED_FIELDS,
            [FieldFilter(FilterOperation.LIKE, "Name", name)],
        )
        if len(routes) == 0:
            raise NoRouteFound("Could not find a ferry route with the specified name")

        result = []

        for route in routes:
            result.append(await ferry_route_from_xml_node(route))

        return result

    async def async_get_next_ferry_stops(
        self,
        from_harbor_name: str,
        to_harnbor_name: str = "",
        after_time: datetime = datetime.now(),
        number_of_stops: int = 1,
    ) -> list[FerryStopModel]:
        """Enable retrieval of next departures."""
        date_as_text = after_time.strftime(DATE_TIME_FORMAT)

        filters: list[FieldFilter | Filter] = [
            FieldFilter(FilterOperation.EQUAL, "FromHarbor.Name", from_harbor_name),
            FieldFilter(
                FilterOperation.GREATER_THAN_EQUAL, "DepartureTime", date_as_text
            ),
        ]
        if to_harnbor_name:
            filters.append(
                FieldFilter(FilterOperation.EQUAL, "ToHarbor.Name", to_harnbor_name)
            )

        sorting = [FieldSort("DepartureTime", SortOrder.ASCENDING)]
        ferry_announcements = await self._api.async_make_request(
            "FerryAnnouncement",
            self.version,
            None,
            FERRY_STOP_REQUIRED_FIELDS,
            filters,
            number_of_stops,
            sorting,
        )

        if len(ferry_announcements) == 0:
            raise NoFerryFound("No FerryAnnouncement found")

        stops = []
        for announcement in ferry_announcements:
            stops.append(await ferry_stop_from_xml_node(announcement))
        return stops

    async def async_get_next_ferry_stop(
        self,
        from_harbor_name: str,
        to_harnbor_name: str = "",
        after_time: datetime = datetime.now(),
    ) -> FerryStopModel:
        """Enable retrieval of next departure."""
        stops = await self.async_get_next_ferry_stops(
            from_harbor_name, to_harnbor_name, after_time, 1
        )
        return stops[0]

    async def async_get_deviation(self, id: str) -> DeviationInfoModel:
        """Retrieve deviation info from Deviation Id."""
        filters: list[FieldFilter | Filter] = [
            FieldFilter(FilterOperation.EQUAL, "Deviation.Id", id)
        ]
        deviations = await self._api.async_make_request(
            "Situation",
            "1.5",
            None,
            DEVIATION_INFO_REQUIRED_FIELDS,
            filters,
        )

        if len(deviations) == 0:
            raise NoDeviationFound("No Deviation found")
        if len(deviations) > 1:
            raise MultipleDeviationsFound("Multiple Deviations found")

        deviation = deviations[0]
        return await deviation_from_xml_node(deviation)



"""Retrieve camera data from Trafikverket API."""

from __future__ import annotations

from pytrafikverket.helpers import camera_from_xml_node
from pytrafikverket.models import CameraInfoModel

from .const import CAMERA_INFO_REQUIRED_FIELDS
from .exceptions import MultipleCamerasFound, NoCameraFound
from .filters import FieldFilter, FilterOperation, OrFilter
from .trafikverket import (
    TrafikverketBase,
)


class TrafikverketCamera(TrafikverketBase):
    """Class used to communicate with trafikverket's camera api."""

    version = "1.0"

    async def async_get_camera(self, search_string: str) -> CameraInfoModel:
        """Retrieve camera from API."""
        cameras = await self._api.async_make_request(
            "Camera",
            self.version,
            None,
            CAMERA_INFO_REQUIRED_FIELDS,
            [
                OrFilter(
                    [
                        FieldFilter(FilterOperation.LIKE, "Name", search_string),
                        FieldFilter(FilterOperation.LIKE, "Location", search_string),
                        FieldFilter(FilterOperation.LIKE, "Id", search_string),
                    ]
                )
            ],
        )
        if len(cameras) == 0:
            raise NoCameraFound("Could not find a camera with the specified name")
        if len(cameras) > 1:
            raise MultipleCamerasFound("Found multiple camera with the specified name")

        return await camera_from_xml_node(cameras[0])

    async def async_get_cameras(self, search_string: str) -> list[CameraInfoModel]:
        """Retrieve multipple cameras from API."""
        cameras = await self._api.async_make_request(
            "Camera",
            self.version,
            None,
            CAMERA_INFO_REQUIRED_FIELDS,
            [
                OrFilter(
                    [
                        FieldFilter(FilterOperation.LIKE, "Name", search_string),
                        FieldFilter(FilterOperation.LIKE, "Location", search_string),
                        FieldFilter(FilterOperation.LIKE, "Id", search_string),
                    ]
                )
            ],
        )
        if len(cameras) == 0:
            raise NoCameraFound("Could not find a camera with the specified name")

        camera_list = []
        for camera in cameras:
            camera_list.append(await camera_from_xml_node(camera))
        return camera_list



"""Retrieve weather station data from Trafikverket API."""

from __future__ import annotations

from .const import WEATHER_REQUIRED_FIELDS
from .exceptions import MultipleWeatherStationsFound, NoWeatherStationFound
from .filters import FieldFilter, FilterOperation
from .helpers import weather_from_xml_node
from .models import WeatherStationInfoModel
from .trafikverket import TrafikverketBase


class TrafikverketWeather(TrafikverketBase):
    """Class used to communicate with trafikverket's weather api."""

    version = "2.1"

    async def async_get_weather(self, location_name: str) -> WeatherStationInfoModel:
        """Retrieve weather from API."""
        weather_stations = await self._api.async_make_request(
            "WeatherMeasurepoint",
            self.version,
            None,
            WEATHER_REQUIRED_FIELDS,
            [FieldFilter(FilterOperation.EQUAL, "Name", location_name)],
        )
        if len(weather_stations) == 0:
            raise NoWeatherStationFound(
                "Could not find a weather station with the specified name"
            )
        if len(weather_stations) > 1:
            raise MultipleWeatherStationsFound(
                "Found multiple weather stations with the specified name"
            )

        return await weather_from_xml_node(weather_stations[0])

    async def async_search_weather_stations(
        self, location_name: str
    ) -> list[WeatherStationInfoModel]:
        """Retrieve weather from API."""
        weather_stations = await self._api.async_make_request(
            "WeatherMeasurepoint",
            self.version,
            None,
            ["Name" "Id"],
            [
                FieldFilter(FilterOperation.LIKE, "Name", location_name),
                FieldFilter(FilterOperation.EQUAL, "Deleted", "false"),
            ],
        )
        if len(weather_stations) == 0:
            raise NoWeatherStationFound(
                "Could not find a weather station with the specified name"
            )

        result = []
        for weather_station in weather_stations:
            result.append(await weather_from_xml_node(weather_station))
        return result



"""Data classes for Trafikverket."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import StrEnum


@dataclass
class CameraInfoModel:
    """Dataclass for Trafikverket Camera."""

    camera_name: str
    camera_id: str
    active: bool | None
    deleted: bool | None
    description: str | None
    direction: str | None
    fullsizephoto: bool | None
    location: str | None
    modified: datetime | None
    phototime: datetime | None
    photourl: str | None
    status: str | None
    camera_type: str | None


@dataclass
class FerryRouteInfoModel:
    """Dataclass for Trafikverket Ferry route info."""

    ferry_route_id: str
    ferry_route_name: str
    short_name: str | None
    route_type: str | None


@dataclass
class DeviationInfoModel:
    """Dataclass for Trafikverket Ferry deviations."""

    deviation_id: str
    header: str | None
    message: str | None
    start_time: datetime | None
    end_time: datetime | None
    icon_id: str | None
    location_desc: str | None


@dataclass
class FerryStopModel:
    """Dataclass for Trafikverket Ferry stop."""

    ferry_stop_id: str
    ferry_stop_name: str | None
    short_name: str | None
    deleted: bool | None
    departure_time: datetime | None
    other_information: list[str] | None
    deviation_id: list[str] | None
    modified_time: datetime | None
    from_harbor_name: str | None
    to_harbor_name: str | None
    type_name: str | None

    def get_state(self) -> FerryStopStatus:
        """Retrieve the state of the departure."""
        if self.deleted:
            return FerryStopStatus.DELETED
        return FerryStopStatus.ON_TIME


class FerryStopStatus(StrEnum):
    """Contain the different ferry stop statuses."""

    ON_TIME = "on_time"
    CANCELED = "canceled"
    DELETED = "deleted"


@dataclass
class StationInfoModel:
    """Dataclass for Trafikverket Train Station info."""

    signature: str
    station_name: str
    advertised: str | None


class TrainStopStatus(StrEnum):
    """Contain the different train stop statuses."""

    ON_TIME = "on_time"
    DELAYED = "delayed"
    CANCELED = "canceled"


@dataclass
class TrainStopModel:
    """Dataclass for Trafikverket Train stop."""

    train_stop_id: str
    canceled: bool | None
    advertised_time_at_location: datetime | None
    estimated_time_at_location: datetime | None
    time_at_location: datetime | None
    other_information: list[str] | None
    deviations: list[str] | None
    modified_time: datetime | None
    product_description: list[str] | None

    def get_state(self) -> TrainStopStatus:
        """Retrieve the state of the departure."""
        if self.canceled:
            return TrainStopStatus.CANCELED
        if (
            self.advertised_time_at_location is not None
            and self.time_at_location is not None
            and self.advertised_time_at_location != self.time_at_location
        ):
            return TrainStopStatus.DELAYED
        if (
            self.advertised_time_at_location is not None
            and self.estimated_time_at_location is not None
            and self.advertised_time_at_location != self.estimated_time_at_location
        ):
            return TrainStopStatus.DELAYED
        return TrainStopStatus.ON_TIME

    def get_delay_time(self) -> timedelta | None:
        """Calculate the delay of a departure."""
        if self.canceled:
            return None
        if (
            self.advertised_time_at_location is not None
            and self.time_at_location is not None
            and self.advertised_time_at_location != self.time_at_location
        ):
            return self.time_at_location - self.advertised_time_at_location
        if (
            self.advertised_time_at_location is not None
            and self.estimated_time_at_location is not None
            and self.advertised_time_at_location != self.estimated_time_at_location
        ):
            return self.estimated_time_at_location - self.advertised_time_at_location
        return None


@dataclass
class WeatherStationInfoModel:
    """Dataclass for Trafikverket Weather info."""

    station_name: str
    station_id: str
    road_temp: float | None  # celsius
    air_temp: float | None  # celsius
    dew_point: float | None  # celsius
    humidity: float | None  # %
    visible_distance: float | None  # meter
    precipitationtype: str | None
    raining: bool | None
    snowing: bool | None
    road_ice: bool | None
    road_ice_depth: float | None  # mm
    road_snow: bool | None
    road_snow_depth: float | None  # mm
    road_water: bool | None
    road_water_depth: float | None  # mm
    road_water_equivalent_depth: float | None  # mm
    winddirection: str | None  # degrees
    wind_height: float | None  # m
    windforce: float | None  # m/s
    windforcemax: float | None  # m/s
    measure_time: datetime | None
    precipitation_amount: float | None  # mm/30min translated to mm/h
    modified_time: datetime | None



"""Pytrafikverket module."""

from .__main__ import main
from .exceptions import (
    InvalidAuthentication,
    MultipleCamerasFound,
    MultipleDeviationsFound,
    MultipleRoutesFound,
    MultipleTrainStationsFound,
    MultipleWeatherStationsFound,
    NoCameraFound,
    NoDeviationFound,
    NoFerryFound,
    NoRouteFound,
    NoTrainAnnouncementFound,
    NoTrainStationFound,
    NoWeatherStationFound,
    UnknownError,
)
from .filters import (
    AndFilter,
    FieldFilter,
    FieldSort,
    Filter,
    FilterOperation,
    OrFilter,
    SortOrder,
)
from .helpers import NodeHelper
from .models import (
    CameraInfoModel,
    DeviationInfoModel,
    FerryRouteInfoModel,
    FerryStopModel,
    FerryStopStatus,
    StationInfoModel,
    TrainStopModel,
    TrainStopStatus,
    WeatherStationInfoModel,
)
from .trafikverket import Trafikverket
from .trafikverket_camera import TrafikverketCamera
from .trafikverket_ferry import TrafikverketFerry
from .trafikverket_train import TrafikverketTrain
from .trafikverket_weather import TrafikverketWeather

__all__ = [
    "AndFilter",
    "CameraInfoModel",
    "DeviationInfoModel",
    "FerryRouteInfoModel",
    "FerryStopModel",
    "FerryStopStatus",
    "FieldFilter",
    "FieldSort",
    "Filter",
    "FilterOperation",
    "InvalidAuthentication",
    "MultipleCamerasFound",
    "MultipleDeviationsFound",
    "MultipleRoutesFound",
    "MultipleTrainStationsFound",
    "MultipleWeatherStationsFound",
    "NoCameraFound",
    "NoDeviationFound",
    "NoFerryFound",
    "NoRouteFound",
    "NoTrainAnnouncementFound",
    "NoTrainStationFound",
    "NoWeatherStationFound",
    "NodeHelper",
    "OrFilter",
    "SortOrder",
    "StationInfoModel",
    "Trafikverket",
    "TrafikverketCamera",
    "TrafikverketFerry",
    "TrafikverketTrain",
    "TrafikverketWeather",
    "TrainStopModel",
    "TrainStopStatus",
    "UnknownError",
    "WeatherStationInfoModel",
]


def run() -> None:
    main()



"""Module for communication with Trafikverket official API."""

from __future__ import annotations

import logging

import aiohttp
from lxml import etree

from .const import API_URL
from .exceptions import InvalidAuthentication, UnknownError
from .filters import FieldFilter, FieldSort, Filter
from .helpers import NodeHelper

LOGGER = logging.getLogger(__name__)


class Trafikverket:
    """Class used to communicate with trafikverket api."""

    def __init__(self, client_session: aiohttp.ClientSession, api_key: str) -> None:
        """Initialize TrafikInfo object."""
        self._client_session = client_session
        self._api_key = api_key

    def _generate_request_data(
        self,
        objecttype: str,
        schemaversion: str,
        namespace: str | None,
        includes: list[str],
        filters: list[Filter],
        limit: int | None = None,
        sorting: list[FieldSort] | None = None,
    ) -> etree._Element:
        root_node = etree.Element("REQUEST")
        login_node = etree.SubElement(root_node, "LOGIN")
        login_node.attrib["authenticationkey"] = self._api_key
        query_node = etree.SubElement(root_node, "QUERY")
        query_node.attrib["objecttype"] = objecttype
        query_node.attrib["schemaversion"] = schemaversion
        if namespace is not None:
            # Only used in some queries
            query_node.attrib["namespace"] = namespace
        if limit is not None:
            query_node.attrib["limit"] = str(limit)
        if sorting is not None and len(sorting) > 0:
            query_node.attrib["orderby"] = ", ".join([fs.to_string() for fs in sorting])
        for include in includes:
            include_node = etree.SubElement(query_node, "INCLUDE")
            include_node.text = include
        filters_node = etree.SubElement(query_node, "FILTER")
        for _filter in filters:
            _filter.generate_node(filters_node)

        return root_node

    async def async_make_request(
        self,
        objecttype: str,
        schemaversion: str,
        namespace: str | None,
        includes: list[str],
        filters: list[Filter | FieldFilter],
        limit: int | None = None,
        sorting: list[FieldSort] | None = None,
    ) -> list[etree._ElementTree]:
        """Send request to trafikverket api and return a element node."""
        request_data = self._generate_request_data(
            objecttype, schemaversion, namespace, includes, filters, limit, sorting
        )
        request_data_text = etree.tostring(request_data, pretty_print=False)
        LOGGER.debug("Sending query with: %s", request_data_text)
        headers = {"content-type": "text/xml"}
        async with self._client_session.post(
            API_URL, data=request_data_text, headers=headers
        ) as response:
            content = await response.text()
            LOGGER.debug("Response received with: %s", content)
            error_nodes = etree.fromstring(content).xpath("/RESPONSE/RESULT/ERROR")
            if len(error_nodes) > 0:
                error_node = error_nodes[0]
                helper = NodeHelper(error_node)
                source = helper.get_text("SOURCE")
                message = helper.get_text("MESSAGE")
                if response.status == 401:
                    raise InvalidAuthentication(
                        f"Source: {source}, message: {message}"
                        ", status: {response.status}"
                    )
                raise UnknownError(
                    f"Source: {source}, message: {message}"
                    ", status: {response.status}"
                )

            result: list[etree._ElementTree] = etree.fromstring(content).xpath(
                "/RESPONSE/RESULT/" + objecttype
            )
            return result


class TrafikverketBase:
    """Base class used to communicate with trafikverket's api."""

    version = "1.0"

    def __init__(self, client_session: aiohttp.ClientSession, api_key: str) -> None:
        """Initialize Trafikverket Base."""
        self._api = Trafikverket(client_session, api_key)
